# Copyright (c) 2019 Red Hat
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import six

from distroinfo import exception
from distroinfo import fetch
from distroinfo import parse


class DistroInfo(object):
    def __init__(self,
                 info_files,
                 fetcher=None,
                 local_info=None,
                 remote_info=None,
                 remote_git_info=None,
                 cache_ttl=3600,
                 cache_base_path=None):
        """
        Specify distroinfo instance to query.

        :param info_files: info file name or a list of file names
        :param fetcher: custom info fetcher,
                        implementation of distroinfo.fetch.InfoFetcher,
                        you can use *_info params instead for builtins
        :param local_info: a shortcut to use LocalInfoFetcher
        :param remote_info: a shortcut to use RemoteRawInfoFetcher
        :param remote_git_info: a shortcut use RemoteGitInfoFetcher
        """
        if fetcher:
            self.fetcher = fetcher
        else:
            # convenience shortcuts to builtin info fetchers
            if local_info:
                self.fetcher = fetch.LocalInfoFetcher(local_info)
            elif remote_info:
                self.fetcher = fetch.RemoteInfoFetcher(
                    remote_info,
                    cache_ttl=cache_ttl,
                    cache_base_path=cache_base_path)
            elif remote_git_info:
                self.fetcher = fetch.RemoteGitInfoFetcher(
                    remote_git_info,
                    cache_ttl=cache_ttl,
                    cache_base_path=cache_base_path)
            else:
                raise exception.InfoFetcherRequired()
        self.fetcher.di_class = DistroInfo
        self.info_files = info_files

        if isinstance(self.info_files, six.string_types):
            # support both a single file and a list
            self.info_files = [self.info_files]

    def get_info(self, apply_tag=None, info_dicts=False):
        """
        Get data from distroinfo instance.

        :param apply_tag: apply supplied tag to info
        :param info_dicts: return packages and releases as dicts
        :return: parsed info metadata
        """
        raw_infos = self.fetcher.fetch(*self.info_files)
        raw_info = parse.merge_infos(*raw_infos, info_dicts=info_dicts)
        info = parse.parse_info(raw_info, apply_tag=apply_tag)
        return info



import pbr.version
import logging


logging.basicConfig(level=logging.WARNING)
logging.getLogger("distroinfo").setLevel(logging.WARNING)

version_info = pbr.version.VersionInfo('distroinfo')
try:
    __version__ = version_info.version_string()
except AttributeError:
    __version__ = None

__all__ = ['__version__']



# Copyright (c) 2019 Red Hat
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from __future__ import print_function
from functools import partial
import re
import six

try:
    from collections.abc import Iterable
except ImportError:
    from collections import Iterable

from distroinfo import exception


def get_release(info, release):
    for rls in info['releases']:
        if rls.get('name') == release:
            return rls
    return None


def get_package(info, name):
    for pkg in info['packages']:
        if pkg['name'] == name:
            return pkg
    return None


def find_package(info, package, strict=False):
    # 1. strict package name matching (openstack-nova)
    pkg = get_package(info, package)
    if pkg:
        return pkg
    # 2. strict project/upstream matching (nova, git://../openstack/nova)
    ps = strip_project_url(package)
    for pkg in info['packages']:
        if 'project' in pkg and pkg['project'].lower() == ps:
            return pkg
        if 'upstream' in pkg:
            upstream = pkg['upstream']
            if strip_project_url(upstream) == ps:
                return pkg
    if strict:
        return None
    # 3. best effort
    is_url = re.search(r'\W', ps)
    psl = ps.lower()
    if is_url:
        for pkg in info['packages']:
            if 'project' in pkg and pkg['project'].lower() in psl:
                return pkg
    for pkg in info['packages']:
        if psl in pkg['name'].lower():
            return pkg
    return None


def find_element(info, needle, info_key='osp_releases'):
    '''Find a matching needle in a custom list of dict'''
    if info_key not in info:
        return None

    for elem in info[info_key]:
        finding = __find_element(elem, needle)
        if finding:
            return elem
    return None


def __find_element(data, needle):
    '''Helper for recursions, used by find_element()'''
    for elem in data:
        if isinstance(data, dict):
            elem = data[elem]
        if needle in elem:
            return data
        if isinstance(elem, list) or isinstance(elem, dict):
            res = __find_element(elem, needle)
            return res
    return None


def filter_pkgs(pkgs, rexen):
    filter_wrapper = partial(_match_pkg, rexen)
    return list(filter(filter_wrapper, pkgs))


def _match_pkg(rexen, pkg):
    for attr, rex in rexen.items():
        val = pkg.get(attr)
        if val is None:
            return False
        if isinstance(val, six.string_types):
            if not re.search(rex, val):
                return False
        elif isinstance(val, Iterable):
            # collection matches if any item of collection matches
            found = False
            for e in val:
                if re.search(rex, e):
                    found = True
                    break
            if not found:
                return False
        else:
            raise exception.InvalidPackageFilter(
                why=("Can only filter strings but '%s' is %s"
                     % (attr, type(rex).__name__)))
    return True


def get_distrepos(info, release, dist=None):
    # release is required now, but this function can be extended to
    # return distrepos for multiple releases if needed
    rls = get_release(info, release)
    if not rls:
        why = 'release not defined in rdoinfo: %s' % release
        raise exception.InvalidQuery(why=why)
    found_dist = False
    distrepos = []
    for repo in rls['repos']:
        if dist and repo['name'] != dist:
            continue
        dr = repo.get('distrepos')
        if dr:
            distrepos.append((release, repo['name'], dr))
            if dist:
                found_dist = True
                break
    if dist and not found_dist:
        why = 'dist not defined in rdoinfo: %s/%s' % (release, dist)
        raise exception.InvalidQuery(why=why)
    if not distrepos:
        why = 'No distrepos information in rdoinfo for %s' % release
        raise exception.InvalidQuery(why=why)

    return distrepos


def tags_diff(info1, info2, tagsname='tags'):
    changedpkgs = []
    for pkg2 in info2["packages"]:
        if pkg2 not in info1["packages"]:
            changedpkgs.append(pkg2)
    diff = []
    for pkg2 in changedpkgs:
        foundpkg = False
        for pkg1 in info1["packages"]:
            if pkg1['project'] == pkg2['project']:
                foundpkg = True
                break
        updated_tags = []
        if foundpkg:
            for tag in pkg2.get(tagsname, {}):
                # use address of this function to differentiate
                # between missing tag and tag: None
                tag1 = pkg1.get(tagsname, {}).get(tag, tags_diff)
                if pkg2[tagsname][tag] != tag1:
                    updated_tags.append(tag)
        else:
            if pkg2.get(tagsname):
                updated_tags = list(pkg2.get(tagsname).keys())
        if updated_tags:
            diff.append((pkg2['name'], updated_tags))
    return diff


def attr_diff(info1, info2, attrname):
    """
    Check for differences in a given attribute between two DistroInfo objects.

       :param DistroInfo info1: initial DistroInfo object to compare with
       :param DistroInfo info2: target DistroInfo object to compare
       :param str attrname: attribute to check for changes

    Returns list of tuples, in the format: ('package-name', 'new-attr-value')
    """
    changedpkgs = []
    for pkg2 in info2["packages"]:
        if pkg2 not in info1["packages"]:
            changedpkgs.append(pkg2)
    diff = []
    for pkg2 in changedpkgs:
        foundpkg = False
        for pkg1 in info1["packages"]:
            if pkg1['project'] == pkg2['project']:
                foundpkg = True
                break
        updated_attrs = None
        if foundpkg:
            attr1 = pkg1.get(attrname)
            if pkg2.get(attrname) != attr1:
                updated_attrs = pkg2.get(attrname)
        else:
            if pkg2.get(attrname):
                updated_attrs = pkg2.get(attrname)
        if updated_attrs:
            diff.append((pkg2['name'], updated_attrs))
    return diff


def strip_project_url(url):
    """strip proto:// | openstack/ prefixes and .git | -distgit suffixes"""
    m = re.match(r'(?:[^:]+://)?(.*)', url)
    if m:
        url = m.group(1)
    if url.endswith('.git'):
        url, _, _ = url.rpartition('.')
    if url.endswith('-distgit'):
        url, _, _ = url.rpartition('-')
    if url.startswith('openstack/'):
        # openstack is always special :-p
        _, _, url = url.partition('/')
    return url



# Copyright (c) 2019 Red Hat
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import collections
import copy
import six

try:
    from collections.abc import Iterable
except ImportError:
    from collections import Iterable

from distroinfo import exception


def parse_info(raw_info, apply_tag=None):
    """
    Parse raw rdoinfo metadata inplace.

    :param raw_info: raw info to parse
    :param apply_tag: tag to apply
    :returns: dictionary containing all packages in rdoinfo
    """
    parse_releases(raw_info)
    parse_packages(raw_info, apply_tag=apply_tag)
    return raw_info


def parse_release_repo(repo, default_branch=None):
    if 'name' not in repo:
        raise exception.MissingRequiredItem(item='repo.name in %s' % repo)
    if 'branch' not in repo:
        if default_branch:
            repo['branch'] = default_branch
        else:
            raise exception.MissingRequiredItem(
                item='repo.branch for repo %s' % repo['name'])
    return repo


def parse_releases(info):
    try:
        releases = info['releases']
    except KeyError:
        raise exception.MissingRequiredSection(section='releases')
    if not isinstance(releases, Iterable):
        raise exception.InvalidInfoFormat(
            msg="'releases' section must be iterable")
    if isinstance(releases, dict):
        releases = releases.values()
    for rls in releases:
        try:
            rls_name = rls['name']
        except KeyError:
            raise exception.MissingRequiredItem(
                item='release.name in %s' % rls)
        try:
            repos = rls['repos']
        except KeyError:
            raise exception.MissingRequiredItem(
                item='release.repos for release %s' % rls_name)
        default_branch = rls.get('branch')
        for repo in repos:
            parse_release_repo(repo, default_branch)
    return releases


def parse_package_configs(info):
    if 'package-default' not in info:
        info['package-default'] = {}
    if 'package-configs' not in info:
        info['package-configs'] = {}
    return info['package-default'], info['package-configs']


def substitute_package(pkg):
    # substitution is very simple, no recursion
    # feel free to extend this as long as you provide proper tests
    new_pkg = copy.copy(pkg)
    for key, val in pkg.items():
        if isinstance(val, six.string_types):
            try:
                new_pkg[key] = val % pkg
            except KeyError:
                raise exception.SubstitutionFailed(txt=val)
    return new_pkg


def parse_package(pkg, info, apply_tag=None):
    pkddefault, pkgconfs = parse_package_configs(info)
    # start with default package config
    parsed_pkg = copy.deepcopy(pkddefault)
    if 'conf' in pkg:
        # apply package configuration template
        conf_id = pkg['conf']
        try:
            conf = pkgconfs[conf_id]
        except KeyError:
            raise exception.UndefinedPackageConfig(conf=conf_id)
        parsed_pkg.update(conf)
    parsed_pkg.update(pkg)
    if apply_tag:
        tags = parsed_pkg.get('tags', {})
        tagdict = tags.get(apply_tag)
        if tagdict:
            parsed_pkg.update(tagdict)
    pkg = substitute_package(parsed_pkg)

    try:
        name = pkg['name']
    except KeyError:
        raise exception.MissingRequiredItem(item='package.name in %s' % pkg)
    if 'project' not in pkg:
        raise exception.MissingRequiredItem(
            item="project for '%s' package" % name)
    try:
        maints = pkg['maintainers']
    except KeyError:
        raise exception.MissingRequiredItem(
            item="maintainers for '%s' package" % name)
    if not maints:
        raise exception.MissingRequiredItem(
            item="at least one maintainer for '%s' package" % name)
    try:
        for maint in maints:
            if '@' not in maint:
                raise exception.InvalidInfoFormat(
                    msg="'%s' doesn't look like maintainer's email." % maint)
    except TypeError:
        raise exception.InvalidInfoFormat(
            msg='package.maintainers must be a list of email addresses')

    return pkg


def _check_for_duplicates(pkg, pkgs):
    for oldpkg in pkgs:
        if pkg['name'] == oldpkg['name']:
            return True
    return False


def parse_packages(info, apply_tag=None):
    try:
        pkgs = info['packages']
    except KeyError:
        raise exception.MissingRequiredSection(section='packages')
    if not isinstance(pkgs, Iterable):
        raise exception.InvalidInfoFormat(
            msg="'packages' section must be iterable")
    if isinstance(pkgs, dict):
        # 'packages' is a dictionary
        info_dicts = True
        parsed_pkgs = collections.OrderedDict()
        pkgs = pkgs.values()
    else:
        # 'packages' is a list
        info_dicts = False
        parsed_pkgs = []
        seen_projects = []
    for pkg in pkgs:
        parsed_pkg = parse_package(pkg, info, apply_tag=apply_tag)
        project = parsed_pkg['project']
        if info_dicts:
            parsed_pkgs[project] = parsed_pkg
        else:
            if project in seen_projects:
                raise exception.DuplicatedProject(prj=parsed_pkg['project'])
            seen_projects.append(project)
            parsed_pkgs.append(parsed_pkg)
    info['packages'] = parsed_pkgs


def _merge(a, b):
    # recursively merge arbitrary data structures
    if isinstance(a, dict) and isinstance(b, dict):
        m = a.copy()
        m.update({k: _merge(a.get(k, None), b[k]) for k in b})
        return m

    if isinstance(a, list) and isinstance(b, list):
        return a + b

    if b is None:
        return a
    return b


def list2dict(_list, key):
    d = collections.OrderedDict()
    for i in _list:
        kv = i.get(key)
        if kv in d:
            d[kv] = _merge(d[kv], i)
        else:
            d[kv] = i
    return d


def info2dicts(info, in_place=False):
    """
    Return info with:

    1) `packages` list replaced by a 'packages' dict indexed by 'project'
    2) `releases` list replaced by a 'releases' dict indexed by 'name'
    """
    if 'packages' not in info and 'releases' not in info:
        return info
    if in_place:
        info_dicts = info
    else:
        info_dicts = info.copy()
    packages = info.get('packages')
    if packages:
        info_dicts['packages'] = list2dict(packages, 'project')
    releases = info.get('releases')
    if releases:
        info_dicts['releases'] = list2dict(releases, 'name')
    return info_dicts


def info2lists(info, in_place=False):

    """
    Return info with:

    1) `packages` dict replaced by a 'packages' list with indexes removed
    2) `releases` dict replaced by a 'releases' list with indexes removed

    info2list(info2dicts(info)) == info
    """
    if 'packages' not in info and 'releases' not in info:
        return info
    if in_place:
        info_lists = info
    else:
        info_lists = info.copy()
    packages = info.get('packages')
    if packages:
        info_lists['packages'] = list(packages.values())
    releases = info.get('releases')
    if releases:
        info_lists['releases'] = list(releases.values())
    return info_lists


def merge_infos(*infos, **kwargs):
    if not infos:
        return {}
    info_dicts = kwargs.get('info_dicts', False)
    in_place = kwargs.get('in_place', True)
    if len(infos) == 1:
        if info_dicts:
            return info2dicts(infos[0], in_place=in_place)
        return infos[0]
    info_dict = info2dicts(infos[0], in_place=in_place)
    for info_next in infos[1:]:
        info_next_dict = info2dicts(info_next, in_place=in_place)
        info_dict = _merge(info_dict, info_next_dict)
    if info_dicts:
        # return info with dicts
        return info_dict
    # convert info back to lists format
    return info2lists(info_dict, in_place=in_place)



from distroinfo.info import DistroInfo
from distroinfo import query
import re

import tests.test_common as common


RDOINFO_GIT_URL = 'https://github.com/redhat-openstack/rdoinfo'
RDOINFO_RAW_URL = ('https://raw.githubusercontent.com/'
                   'redhat-openstack/rdoinfo/master/')


def test_rdoinfo_remote_fetch(tmpdir):
    log_stream = common.capture_distroinfo_logger()
    remote_di = DistroInfo('rdo-full.yml',
                           remote_info=RDOINFO_RAW_URL,
                           cache_base_path=str(tmpdir))
    # first fetch
    remote_info = remote_di.get_info()
    out = log_stream.getvalue()
    assert 'Fetching remote file:' in out
    # NOTE(jpena): with the current rdoinfo structure, we import tags.yml
    # twice, so it would fail the assertion
    # assert 'cached' not in out
    common.assert_rdoinfo_full(remote_info)
    # second fetch should be cached
    remote_info = remote_di.get_info()
    out = log_stream.getvalue()
    assert re.search(r'Using \d+ s old cached version of [^\n]+$', out)
    common.assert_rdoinfo_full(remote_info)
    # also load and parse the local repo copy (cache) using local fetcher
    cached_di = DistroInfo('rdo-full.yml',
                           local_info=remote_di.fetcher.cache_path)
    cached_info = cached_di.get_info()
    common.assert_rdoinfo_full(cached_info)
    # make sure results from both fetchers are identical
    assert remote_info == cached_info


def test_rdoinfo_git_fetch(tmpdir):
    log_stream = common.capture_distroinfo_logger()
    git_di = DistroInfo('rdo-full.yml',
                        remote_git_info=RDOINFO_GIT_URL,
                        cache_base_path=str(tmpdir))
    # first fetch should clone the git repo
    git_info = git_di.get_info()
    out = log_stream.getvalue()
    assert u"Cloning git repo" in out
    common.assert_rdoinfo_full(git_info)
    # second fetch with the same instance uses the already synced repo
    # without additional checks
    log_stream = common.capture_distroinfo_logger()
    git_info = git_di.get_info()
    assert log_stream.getvalue() == ''
    common.assert_rdoinfo_full(git_info)
    # a followup fetch with a new instance should reuse previously synced repo
    # after ttl checks
    git_di = DistroInfo('rdo-full.yml',
                        remote_git_info=RDOINFO_GIT_URL,
                        cache_base_path=str(tmpdir))
    git_info = git_di.get_info()
    out = log_stream.getvalue()
    assert u"git repo is fresh" in out
    assert u"Cloning git repo" not in out
    assert u"Fetching git repo" not in out
    assert u"git resp is too old" not in out

    common.assert_rdoinfo_full(git_info)
    # also load and parse the local repo copy (cache) using local fetcher
    cached_di = DistroInfo('rdo-full.yml',
                           local_info=git_di.fetcher.cache_path)
    cached_info = cached_di.get_info()
    common.assert_rdoinfo_full(cached_info)
    # make sure results from both fetchers are identical
    assert git_info == cached_info
